import os

from office365.sharepoint.client_context import ClientContext
from office365.runtime.auth.client_credential import ClientCredential
from office365.runtime.auth.user_credential import UserCredential

class sharepoint:
    """
    Allows interaction with SharePoint environment.
    Need to obtain credentials: https://learn.microsoft.com/en-us/sharepoint/dev/solution-guidance/security-apponly-azureacs
    Either the pair client_id - client_secret is required, either the pair username - user_password

    Parameters
    ----------
    site_url : str
        SharePoint site.
    client_id : str (optional)
        Client id.
    client_secret : str (optional)
        Client secret.
    username : str (optional)
        User name of a SharePoint user account.
    user_password : str (optional)
        User password of a SharePoint user account.
    """

    def __init__(self, site_url, client_id=None, client_secret=None, username=None, user_password=None):

        if client_id is not None:
            credentials = ClientCredential(client_id, client_secret)
        elif username is not None:
            credentials = UserCredential(username, user_password)

        self.env = ClientContext(site_url).with_credentials(credentials)

    def download(self, input_path, output_path):
        """
        Download a file from SharePoint.
        
        Parameters
        ----------
        input_path : str
            Sharepoint path to the file to be downloaded
        output_path : str
            Local path to store the downloaded file
        """

        with open(output_path, "wb") as local_file:
           (self.env
            .web
            .get_file_by_server_relative_path(input_path)
            .download(local_file)
            .execute_query()
                     )
           
    def upload(self, input_path, output_path):
        """
        Upload a file into SharePoint.
        
        Parameters
        ----------
        input_path : str
            Local path to the file to be uploaded
        output_path : str
            SharePoint path to store the uploaded file
        """
        
        with open(input_path, 'rb') as content_file:
            file_content = content_file.read()

        dir, name = os.path.split(output_path)
        self.env.web.get_folder_by_server_relative_url(dir).upload_file(name, file_content).execute_query()
           
    def list_files(self, folder):
        """
        List the files in a SharePoint folder
        
        Parameters
        ----------
        folder : str
            Sharepoint access path for listing files
        """

        root_folder = self.env.web.get_folder_by_server_relative_path(folder)
        root_folder.expand(["Files", "Folders"]).get().execute_query()
        files = [file.properties['ServerRelativeUrl'].split('/')[-1] for file in root_folder.files]

        return files
    
    def delete_file(self, file_path):
        """
        Delete a file
        
        Parameters
        ----------
        file_path : str
            Sharepoint path of the file to be deleted
        """

        path_env = self.env.web.get_file_by_server_relative_url(file_path)
        path_env.delete_object().execute_query()


import os

import fsspec

from google.cloud import storage, bigquery

from .file_manager import FileManager

class gcloud:
    """
    Allows interaction with Google Cloud environment.

    Parameters
    ----------
    project_id : str
        The ID of the Google Cloud project.
    credential_path : str
        The path to the Google Cloud credentials file. Default is None.
    GCS_path : str
        The base path of the Google Cloud Storage.
    extra_loader_config : dict
        Extra configuration for file loaders.
    extra_saver_config : dict
        Extra configuration for file savers.
    """

    def __init__(self, project_id, credential_path=None, GCS_path=None,
                 extra_loader_config=None, extra_saver_config=None):

        self.GCS = GCS(project_id=project_id, GCS_path=GCS_path, credential_path=credential_path,
                       extra_loader_config=extra_loader_config, extra_saver_config=extra_saver_config)
        self.BQ = BQ(project_id=project_id, credential_path=credential_path)

class GCS:
    """
    Allows interaction with Google Cloud Storage environment.

    Parameters
    ----------
    project_id : str
        The ID of the Google Cloud project.
    credential_path : str
        The path to the Google Cloud credentials file. Default is None.
    GCS_path : str
        The base path on GCS. Default is None.
    extra_loader_config : dict
        Extra configuration for file loaders.
    extra_saver_config : dict
        Extra configuration for file savers.
    """

    def __init__(self, project_id, credential_path=None, GCS_path=None, extra_loader_config=None, extra_saver_config=None):
    
        self.project_id = project_id
        self.GCS_path = GCS_path
        self.file_manager = FileManager(extra_loader_config, extra_saver_config)
        self.credential_path = credential_path

        if credential_path is not None:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credential_path

    def get_blob(self, path):
        """
        Retrieves a blob from the specified path in Google Cloud Storage.

        Parameters
        ----------
        path : str
            path to blob
        """

        full_path = self.GCS_path + path
        bucket_name, path = full_path[5:].split('/', 1)
        blob = storage.Client(project=self.project_id).bucket(bucket_name).blob(path)
        return blob

    def load(self, path, **kwargs):
        """
        Load a file from GCS
        To learn more about the extensions supported by default, refer to the documentation : https://antoinepinto.gitbook.io/easy-environment/
        To integrate other extensions into the tool, see documentation "Customise supported formats": https://antoinepinto.gitbook.io/easy-environment/extra/customise-supported-formats
        
        Parameters
        ----------
        path : str
            path to load from
        """

        extension = path.split('.')[-1]

        if extension not in self.file_manager.loader_config:
            error_message = (
                f"Extension '{extension}' is not currently supported through Easy Environment. You can\n"
                "customise supported extensions (see documentation https://antoinepinto.gitbook.io/easy-environment/extra/customise-supported-formats)"
            )
            raise ValueError(error_message)

        full_path = self.GCS_path + path
        loader, mode = self.file_manager.loader_config[extension]
        with fsspec.open(full_path, mode, token=self.credential_path) as f:
            return loader(f, **kwargs)

    def save(self, obj, path, **kwargs):
        """
        Save a file to GCS
        To learn more about the extensions supported by default, refer to the documentation : https://antoinepinto.gitbook.io/easy-environment/
        To integrate other extensions into the tool, see documentation "Customise supported formats": https://antoinepinto.gitbook.io/easy-environment/extra/customise-supported-formats

        Parameters
        ----------
        obj
            object to save
        path : str
            path to save to
        """

        extension = path.split('.')[-1]

        if extension not in self.file_manager.saver_config:
            error_message = (
                f"Extension '{extension}' is not currently supported through Easy Environment. You can\n"
                "customise supported extensions (see documentation https://antoinepinto.gitbook.io/easy-environment/extra/customise-supported-formats)"
            )
            raise ValueError(error_message)

        if extension in ['png', 'jpg']:
            error_message = (
                f"Extension '{extension}' is not currently supported for saving in Google Cloud Storage\n"
                "through Easy Environment."
            )
            raise ValueError(error_message)

        full_path = self.GCS_path + path
        saver, mode = self.file_manager.saver_config[extension]
        with fsspec.open(full_path, mode, token=self.credential_path) as f:
            return saver(obj, f, **kwargs)

    def list_files(self, path):
        """
        List files into a specific folder.
        
        Parameters
        ----------
        path : str
            path to list files
        """

        full_path = self.GCS_path + path
        bucket_name, path = full_path[5:].split('/', 1)
        client = storage.Client(project=self.project_id).bucket(bucket_name)
        files = [blob.name for blob in client.list_blobs(prefix=path)]

        return files
    
    def download(self, path, output_path):
        """
        Download a file from the specified path on Google Cloud Storage.
        
        Parameters
        ----------
        path : str
            path to the file to download
        output_path : str
            path to store the file
        """

        blob = self.get_blob(path)
        blob.download_to_filename(output_path)
    
class BQ:
    """
    Allows interaction with Google Cloud Big Query environment.

    Parameters
    ----------
    project_id : str
        The ID of the Google Cloud project.
    credential_path : str
        The path to the Google Cloud credentials file. Default is None.
    """

    def __init__(self, project_id, credential_path=None):

        self.project_id = project_id

        if credential_path is not None:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credential_path

    def load(self, path):
        """
        Load an entire Big Query table into Python.
        
        Parameters
        ----------
        path : str
            path representing the data set and the name of the table (ex : "mydata.mytable")
        """

        query = f"SELECT * FROM `{path}`"
        client = bigquery.Client(project=self.project_id)
        return client.query(query).result().to_dataframe()

    def write(self, obj, path, schema=None):
        """
        Write an entire Python dataframe into Big Query.
        
        Parameters
        ----------
        obj : pandas.DataFrame
            table to save
        path : str
            path representing the data set and the name of the table (ex : "mydata.mytable")
        schema : list (optional)
            schema of the table. If not specified, a schema is generated based on mapping
            Format: list of dictionnaries (see Documentation)
        """

        client = bigquery.Client(project=self.project_id)

        job_config = bigquery.LoadJobConfig(
            autodetect=True,
            source_format=bigquery.SourceFormat.PARQUET,
            write_disposition='WRITE_TRUNCATE'
        )

        if schema is not None:
            job_config.schema = schema

        client.load_table_from_dataframe(obj, path, job_config=job_config)

    def append(self, obj, path):
        """
        Append an existing Big Query table.
        
        Parameters
        ----------
        obj : pandas.DataFrame
            table to append
        path : str
            path representing the data set and the name of the table (ex : "mydata.mytable")
        """

        job_config = bigquery.LoadJobConfig(
            autodetect=True,
            source_format=bigquery.SourceFormat.PARQUET,
            write_disposition='WRITE_APPEND'
        )

        client = bigquery.Client(project=self.project_id)
        client.load_table_from_dataframe(obj, path, job_config=job_config)
 
    def query(self, query):
        """
        Execute Big Query query.
        
        Parameters
        ----------
        query : str
            query to execute
        """

        client = bigquery.Client(project=self.project_id)
        return client.query(query)


import os

import fsspec

from .file_manager import FileManager

class Disk:
    """
    Allows interaction with local environment.

    Parameters
    ----------
    local_path : str
        The path from which interactions with the local environment will be performed.
    extra_loader_config : dict
        Extra configuration for file loaders. Default is None.
    extra_saver_config :dict
        Extra configuration for file savers. Default is None.
    """

    def __init__(self, root_path, extra_loader_config=None, extra_saver_config=None):
        self.root_path = root_path
        self.file_manager = FileManager(extra_loader_config, extra_saver_config)

    def load(self, path, **kwargs):
        """
        Load a file.
        To learn more about the extensions supported by default, refer to the documentation : https://antoinepinto.gitbook.io/easy-environment/
        To integrate other extensions into the tool, see documentation "Customise supported formats": https://antoinepinto.gitbook.io/easy-environment/extra/customise-supported-formats

        Parameters
        ----------
        path : str
            path to load from.
        """

        load_path = os.path.join(self.root_path, path)
        extension = path.split('.')[-1]

        if extension not in self.file_manager.loader_config:
            error_message = (
                f"Extension '{extension}' is not currently supported through Easy Environment. You can\n"
                "customise supported extensions (see documentation https://antoinepinto.gitbook.io/easy-environment/extra/customise-supported-formats)"
            )
            raise ValueError(error_message)

        loader, mode = self.file_manager.loader_config[extension]

        with fsspec.open(load_path, mode) as f:
            return loader(f, **kwargs)

    def save(self, obj, path, **kwargs):
        """
        Save a file
        To learn more about the extensions supported by default, refer to the documentation : https://antoinepinto.gitbook.io/easy-environment/
        To integrate other extensions into the tool, see documentation "Customise supported formats": https://antoinepinto.gitbook.io/easy-environment/extra/customise-supported-formats

        Parameters
        ----------
        obj
            object to save.
        path : str
            path to save to.
        """

        save_path = os.path.join(self.root_path, path)
        extension = path.split('.')[-1]

        if extension not in self.file_manager.saver_config:
            error_message = (
                f"Extension '{extension}' is not currently supported through Easy Environment. You can\n"
                "customise supported extensions (see documentation https://antoinepinto.gitbook.io/easy-environment/extra/customise-supported-formats)"
            )
            raise ValueError(error_message)

        saver, mode = self.file_manager.saver_config[extension]

        with fsspec.open(save_path, mode) as f:
            saver(obj, f, **kwargs)

    def clear_folder(self, path):
        """
        Clear a folder
        
        Parameters
        ----------
        path : str
            path at which to delete all files.
        """
        folder_path = os.path.join(self.root_path, path)
        for filename in os.listdir(folder_path):
            file_path = os.path.join(folder_path, filename)
            os.remove(file_path)


with open("README.md", "r") as f:
    long_description = f.read()

from setuptools import setup, find_packages

setup(
    name='easyenvi',
    version='1.0.4',
    description='Easy-to-use functionality for managing files and data in different environments',
    author='Antoine PINTO',
    author_email='antoine.pinto1@outlook.fr',
    license='MIT',
    license_file='LICENSE',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/AntoinePinto/easy-environment',
    project_urls={
        'Documentation': 'https://antoinepinto.gitbook.io/easy-environment/',
        'Source Code': 'https://github.com/AntoinePinto/easy-environment',
    },
    packages=find_packages(),
    install_requires=[
    ],
)


class FileManager:
    """
    Manages loading and saving files using various formats.
    
    Parameters
    ----------
    extra_loader_config : dict
        Extra configuration for file loaders.
    extra_saver_config : dict
        Extra configuration for file savers.
    """

    def __init__(self, extra_loader_config=None, extra_saver_config=None):

        self.loader_config = {
            'csv':      (csv_loader,        'rb'),
            'docx':     (docx_loader,       'rb'),
            'jpg':      (jpg_loader,        'rb'),
            'json':     (json_loader,       'rt'),
            'md':       (md_loader,         'rt'),
            'parquet':  (parquet_loader,    'rb'),
            'pdf':      (pdf_loader,        'rb'),
            'pickle':   (pickle_loader,     'rb'),
            'png':      (png_loader,        'rb'),
            'pptx':     (pptx_loader,       'rb'),
            'sql':      (sql_loader,        'rt'),
            'toml':     (toml_loader,       'rt'),
            'txt':      (txt_loader,        'rt'),
            'xlsx':     (xlsx_loader,       'rb'),
            'xml':      (xml_loader,        'rb'),
            'yaml':     (yaml_loader,       'rt'),
            'yml':      (yaml_loader,       'rt')
        }
        
        self.saver_config = {
            'csv':      (csv_saver,         'wb'),
            'docx':     (docx_saver,        'wb'),
            'jpg':      (jpg_saver,         'wb'),
            'json':     (json_saver,        'wt'),
            'md':       (md_saver,          'wt'),
            'parquet':  (parquet_saver,     'wb'),
            'pdf':      (pdf_saver,         'wb'),
            'pickle':   (pickle_saver,      'wb'),
            'png':      (png_saver,         'wb'),
            'pptx':     (pptx_saver,        'wb'),
            'sql':      (sql_saver,         'wt'),
            'toml':     (toml_saver,        'wt'),
            'txt':      (txt_saver,         'wt'),
            'xlsx':     (xlsx_saver,        'wb'),
            'xml':      (xml_saver,         'wb'),
            'yaml':     (yaml_saver,        'wt'),
            'yml':      (yaml_saver,        'wt')
        }

        if extra_loader_config is not None:
            self.loader_config.update(extra_loader_config)

        if extra_saver_config is not None:
            self.saver_config.update(extra_saver_config)

# CSV
def csv_loader(path, **kwargs):
    import pandas as pd
    return pd.read_csv(path, **kwargs)

def csv_saver(obj, path, **kwargs):
    obj.to_csv(path, **kwargs)

# DOCX
def docx_loader(path):
    from docx import Document
    return Document(path)

def docx_saver(obj, path):
    obj.save(path)

# JPG
def jpg_loader(path, **kwargs):
    from PIL import Image
    return Image.open(path, **kwargs).copy()

def jpg_saver(obj, path, **kwargs):
    obj.save(path, **kwargs)

# JSON
def json_loader(path, **kwargs):
    import json
    return json.load(path, **kwargs)

def json_saver(obj, path, **kwargs):
    import json
    json.dump(obj, path, **kwargs)

# MD
def md_loader(path, **kwargs):
    return path.read(**kwargs)

def md_saver(obj, path, **kwargs):
    path.write(obj, **kwargs)

# PARQUET
def parquet_loader(path, **kwargs):
    import pandas as pd
    return pd.read_parquet(path, **kwargs)

def parquet_saver(obj, path, **kwargs):
    obj.to_parquet(path, **kwargs)

# PDF
def pdf_loader(path):
    from PyPDF2 import PdfReader
    # return PdfReader(path)
    import copy
    return copy.deepcopy(PdfReader(path))

def pdf_saver(obj, path):
    from PyPDF2 import PdfWriter
    output = PdfWriter()
    for page in obj.pages:
        output.add_page(page)
    
    output.write(path)

# PICKLE
def pickle_loader(path, **kwargs):
    import pickle
    return pickle.load(path, **kwargs)

def pickle_saver(obj, path, **kwargs):
    import pickle
    pickle.dump(obj, path, protocol=pickle.HIGHEST_PROTOCOL, **kwargs)

# PPTX
def pptx_loader(path):
    from pptx import Presentation
    return Presentation(path)

def pptx_saver(obj, path):
    obj.save(path)

# PNG
def png_loader(path, **kwargs):
    from PIL import Image
    return Image.open(path, **kwargs).copy()

def png_saver(obj, path, **kwargs):
    obj.save(path, **kwargs)

# SQL
def sql_loader(path, **kwargs):
    return path.read(**kwargs)

def sql_saver(obj, path, **kwargs):
    path.write(obj, **kwargs)

# TOML
def toml_loader(path, **kwargs):
    import toml
    return toml.load(path, **kwargs)

def toml_saver(obj, path, **kwargs):
    import toml
    toml.dump(obj, path, **kwargs)

# TXT
def txt_loader(path, **kwargs):
    return path.read(**kwargs)

def txt_saver(obj, path, **kwargs):
    path.write(obj, **kwargs)

# XML
def xml_loader(path, **kwargs):
    import xml.etree.ElementTree as ET
    tree = ET.parse(path, **kwargs)
    root = tree.getroot()
    return root

def xml_saver(obj, path, **kwargs):
    import xml.etree.ElementTree as ET
    tree = ET.ElementTree(obj)
    tree.write(path, **kwargs)

# XLSX
def xlsx_loader(path, **kwargs):
    import pandas as pd
    return pd.read_excel(path, **kwargs)

def xlsx_saver(obj, path, **kwargs):
    obj.to_excel(path, **kwargs)

# YAML
def yaml_loader(path, **kwargs):
    import yaml
    return yaml.safe_load(path, **kwargs)

def yaml_saver(obj, path, **kwargs):
    import yaml
    yaml.dump(obj, path, **kwargs)